import BeaconTau
from bokeh.plotting import figure, gridplot, curdoc
from bokeh.models import ColumnDataSource

import bokeh.palettes as bpal
from bokeh.layouts import row, column
from bokeh.models.widgets import Button, Toggle, Div

import numpy as np

from bokeh.server.server import Server
from bokeh.application import Application
from bokeh.application.handlers.function import FunctionHandler

from datetime import datetime
from time import sleep

from scipy import signal
from math import log10

class Flame:
    """(Facility for Live Action Monitoring of Events) The online event viewer for BeaconTau"""
    def __init__(self):
        self.run = 99
        self.event_lines = {}
        self.event_plots = {}

        self.current_entry = 0
        self.play_state = 0
        self.rev_state = 0

        self.dt = 2.0 #e-9 #ns 500 Mega Samples Per Second

        self.reader = BeaconTau.RunReader(self.run, '../../../data')
        self.title_div = Div(text = "BEACON FLAME (Facility for Live Action Monitoring of Events)")
        self.header_div = Div(text = self.get_header_text())
        self.b_play = None
        self.b_rev = None
        self.freq_domain = False

        self.setup_graphics()

    def get_header_text(self):
        h = self.reader.headers[self.current_entry]
        time_text = datetime.fromtimestamp(h.readout_time[0]).strftime('%Y-%m-%d %H:%M:%S')
        text = "Run "  + str(self.run) + " event " + str(h.event_number) + " (" + time_text + ")"
        return text


    def create_event_plot(self):
        e = self.reader.events[self.current_entry]
        h = self.reader.headers[self.current_entry]
        board_plots = []

        for board, board_channels in enumerate(e.data):
            plots = []
            if board not in self.event_lines:
                self.event_lines[board] = {}
                self.event_plots[board] = {}

            for channel, waveform in enumerate(board_channels):
                chan_title = 'Channel ' + str(channel+1)
                p = figure(title = chan_title, plot_width = 400, plot_height = 200, output_backend="webgl", tools="", toolbar_location = None, logo = None)
                source = ColumnDataSource({'x' : [], str(channel) : []})
                self.event_plots[board][channel] = p
                self.event_lines[board][channel] = p.line(x = 'x', y = str(channel), source = source, color = bpal.Category20[len(board_channels)*2][channel*2])

                if channel > 0:
                    p.y_range = plots[0].y_range
                plots.append(p)


            half_n = int(len(plots)/2)
            top_row = plots[:half_n]
            bottom_row = plots[half_n:]
            p = gridplot(top_row, bottom_row)
            board_plots.append(p)

        self.update_events(self.current_entry)
        return column(board_plots)

    def toggle_domain(self, state):
        self.freq_domain = state
        self.update_events(self.current_entry)

    def toggle_play(self, state):
        self.play_state = state
        if self.play_state > 0:
            self.b_play.label = 'Pause'
            self.play_callback_id = curdoc().add_periodic_callback(self.next_event, 1000)
        else:
            self.b_play.label = 'Play'
            curdoc().remove_periodic_callback(self.play_callback_id)

    def toggle_reverse(self, state):
        self.rev_state = state
        if self.rev_state > 0:
            self.b_rev.label = 'Pause'
            self.rev_callback_id = curdoc().add_periodic_callback(self.previous_event, 1000)
        else:
            self.b_rev.label = 'Reverse'
            curdoc().remove_periodic_callback(self.rev_callback_id)


    def next_event(self):
        entry = self.current_entry + 1
        self.update_events(entry)


    def previous_event(self):
        entry = self.current_entry - 1
        self.update_events(entry)


    def update_events(self, entry):
        if entry < 0:
            entry = entry + len(self.reader.events)
        elif entry >= len(self.reader.events):
            entry = entry - len(self.reader.events)

        self.current_entry = entry

        e = self.reader.events[self.current_entry]
        self.event_number = self.reader.headers[self.current_entry].event_number
        self.header_div.text = self.get_header_text()

        for board, board_channels in enumerate(e.data):
            channel_data = {}
            for channel, waveform in enumerate(board_channels):
                w = waveform[:e.buffer_length]
                if self.freq_domain == False:
                    if 'x' not in channel_data:
                        t = [self.dt*i for i in range(len(w))]
                        channel_data['x'] = t
                    channel_data[str(channel)] = w
                else:
                    df = 1e3/(len(w)*self.dt) # 1e3 GHz -> MHz
                    freqs, Pxx_den = signal.periodogram(w, df)
                    Pxx_den = [10*log10(y) if y > 0 else -40 for y in Pxx_den]
                    if 'x' not in channel_data:
                        channel_data['x'] = freqs[1:]
                    channel_data[str(channel)] = Pxx_den[1:]


            curdoc().hold()
            for channel in range(len(board_channels)):
                if channel in self.event_lines[board]:
                    self.event_lines[board][channel].data_source.data = channel_data

                    if self.freq_domain:
                        self.event_plots[board][channel].xaxis.axis_label = 'Frequency (MHz)'
                        self.event_plots[board][channel].yaxis.axis_label = 'PSD (???)'
                    else:
                        self.event_plots[board][channel].xaxis.axis_label = 'Time (ns)'
                        self.event_plots[board][channel].yaxis.axis_label = 'Amplitude (mV)'
            curdoc().unhold()

    def plot_thresholds(self):
        p = figure(title = 'Thresholds', plot_width=1600, plot_height=400, toolbar_location = None,  tools = "", logo = None)
        long_cols = bpal.Category20[20] + bpal.Category20b[20] + bpal.Category20c[20]
        rot = [st.readout_time for st in self.reader.statuses]
        threshs = [st.trigger_thresholds for st in self.reader.statuses]
        threshs2 = list(map(list, zip(*threshs)))
        for i, thresh in enumerate(threshs2):
            leg_text = 'Beam ' + str(i+1)
            p.line(rot,thresh,color=long_cols[i], legend = leg_text, line_width=2)

        p.legend.click_policy="hide"

        return p

    def setup_graphics(self):

        p1 = self.create_event_plot()
        p2 = self.plot_thresholds()

        b_next = Button(label="Next", button_type="success", width = 400)
        b_next.on_click(self.next_event)

        b_prev = Button(label="Prev", button_type="success", width = 400)
        b_prev.on_click(self.previous_event)

        self.b_play = Toggle(label = "Play", button_type = "success",  width = 400)
        self.b_play.on_click(self.toggle_play)

        self.b_rev = Toggle(label = "Rev", button_type = "success",  width = 400)
        self.b_rev.on_click(self.toggle_reverse)

        self.b_domain = Toggle(label = 'Freq', button_type="success", width = 400)
        self.b_domain.on_click(self.toggle_domain)
        #self.b_average = Toggle(label = 'Average', button_type="success", width = 400)

        b1 = row([self.b_play, self.b_rev])
        b2 = row([b_prev, b_next])
        b3 = row([self.b_domain])

        curdoc().title = 'BEACON FLAME'
        curdoc().add_root(self.title_div)
        curdoc().add_root(self.header_div)
        curdoc().add_root(b1)
        curdoc().add_root(b2)
        curdoc().add_root(b3)
        curdoc().add_root(p1)
        curdoc().add_root(p2)

if __name__ == '__main__':
    pass
elif 'bk_script' in __name__:
    # Then this was called by bokeh serve
    # So we create an instance of the web viewer
    f = Flame()



from _BeaconTau import * # Load the pybind11 stuff defined in _BeaconTau as if it's in BeaconTau

from .EventAnalyzer import EventAnalyzer
from .RunAnalyzer import RunAnalyzer
from .DataDirectory import DataDirectory

name = "BeaconTau"

def main():
    d = DataDirectory()

    for r in d:
        for entry in range(2):
            e = r.get_entry(entry)
            e.plot()
        r.draw('trigger_thresholds')

# plt.ion()
# plt.show()
if __name__ == '__main__':
    main()




from _BeaconTau import * # This imports the c++ defined code from BeaconTau.cpp
import os
from .RunAnalyzer import RunAnalyzer


class DataDirectory():
    """
    High level interface to the BEACON data directory structures.
    Allows iteration of RunAnalyzers over all runs in the directory

    First looks in an optional data_dir string.
    If one is not provided, then it looks for a BEACON_DATA_DIR environment variable.
    If there isn't one it looks in the current directory.
    If it doesn't find any run subfolders in any of those places, it gives up.
    """
    def __init__(self, data_dir = None):

        self.data_dir = data_dir

        if self.data_dir == None:
            try:
                self.data_dir = os.environ['BEACON_DATA_DIR']
                print('Found ' + self.data_dir + ' from BEACON_DATA_DIR')
            except:
                print('No BEACON_DATA_DIR environment variable, setting to current directory')
                self.data_dir = os.getcwd()

        self.run_dirs = [d for d in os.listdir(self.data_dir) if 'run' in d]

        if len(self.run_dirs) == 0:
            raise ValueError("Couldn't find any runs under " + self.data_dir)

        self.runs = sorted([int(x.strip('run')) for x in self.run_dirs])
        self._i =  0

    def __repr__(self):
        return '<BeaconTau.DataDirectory at ' + str(self.data_dir) + ' containing runs ' + str(self.runs) + '>'

    def __iter__(self):
        return self

    def __next__(self):
        if self._i < len(self.runs):
            self._i += 1 # For next call to __next__
            return RunAnalyzer(self.runs[self._i - 1], self.data_dir)
        else:
            raise StopIteration

    def run(self, run_number):
        run_index = -1
        try:
            run_index = self.runs.index(run_number)
        except:
            raise ValueError('No run ' +  str(run_number) + ' in ' + self.data_dir + ', available runs are ' + str(self.runs))

        return RunAnalyzer(run_number, self.data_dir)



from _BeaconTau import * # This imports the c++ defined code from BeaconTau.cpp
import matplotlib.pyplot as plt
from scipy import signal
from math import floor, log10

class EventAnalyzer():
    """
    Do some useful things with the raw Event and Header classes defined in BeaconTau.cpp
    The magic of pybind11 only gets us so far: it's hard to extend the same c++ classes in python.
    Inheritance is also pain to implement (and would be inefficient to copy data to the derived class).
    So instead this we have this EventAnalyzer class that has the event and header as members, and does useful things with them.
    """

    def __init__(self, header, event):
        if header.event_number != event.event_number:
            raise ValueError('Mismatched header and event!')
        self.header = header
        self.event = event

        self.time_array = None # Constructed on demand
        self.freq_array = None # Constructed on demand
        self.dt = 2.0 # Nano seconds
        self.df = 1e3/(self.event.buffer_length*self.dt) # 1e3 GHz -> MHz


    def __repr__(self):
        return '<BeaconTau.EventAnalyzer for event ' + str(self.event.event_number) + '>'

    def channel(self, chan_index, board_index = 0):
        # The primary method of accessing the channel data.
        # Automatically trims the buffer to the correct length.
        return self.event.data[board_index][chan_index][:self.event.buffer_length]

    def channel_psd(self, chan_index, board_index = 0):
        freqs, psd = signal.periodogram(self.channel(chan_index), self.df)
        return psd

    def channel_psd_db(self, chan_index, board_index = 0):
        psd = self.channel_psd(chan_index, board_index = board_index)
        psd_db = [log10(p) if p > 0 else 0 for p in psd]
        return psd_db

    def times(self):
        # Dynamically constructs the time array for the event
        if self.time_array is None:
            self.time_array = [self.dt*i for i in range(self.event.buffer_length)]
        return self.time_array

    def freqs(self):
        if self.freq_array is None:
            nf = floor(self.event.buffer_length/2) + 1
            self.freq_array = [self.df*j for j in range(nf)]
        return self.freq_array

    def plot(self, n_rows = 2, show = False, axes = None, freq_domain = False, log_scale = True):
        # Draw the event in Matplotlib
        for board in self.event.data:
            n_cols = int(len(board)/n_rows)

            # TODO add check on axes length if extenal set of axes passed in
            # TODO for multiple boards, maybe that's overkill...
            if axes is None:
                fig, axes = plt.subplots(n_rows, n_cols, sharey = True, sharex = True)
                mng = plt.get_current_fig_manager()
                mng.resize(*mng.window.maxsize())
            for chan in range(len(board)):
                axes.flat[chan].set_title('Channel ' + str(chan+1))
                color = 'C' + str(chan) # matplotlib accepts strings beginning with a capital C for colors
                if freq_domain is False:
                    axes.flat[chan].plot(self.times(),  self.channel(chan), color)
                    axes.flat[chan].set_xlabel('Time (ns)')
                    axes.flat[chan].set_ylabel('Amplitude (mV)')
                else:
                    axes.flat[chan].set_xlabel('Freq (MHz)')
                    if log_scale is True:
                        axes.flat[chan].plot(self.freqs(), self.channel_psd_db(chan), color)
                        axes.flat[chan].set_ylabel('PSD (dB)')
                    else:
                        axes.flat[chan].plot(self.freqs(), self.channel_psd(chan), color)
                        axes.flat[chan].set_ylabel('PSD (mV^{2} / MHz)')
            plt.suptitle('Event ' + str(self.event.event_number))
        if show is True:
            plt.show()



from _BeaconTau import * # This imports the c++ defined code from BeaconTau.cpp
from .EventAnalyzer import EventAnalyzer
import matplotlib.pyplot as plt

class RunAnalyzer():
    """
    Class for examining data in a run.
    Can plot/scan any attribute in the Status/Header/Event files.
    Can pull up individual events by event_number or entry.
    Wraps the file reading action from FileReader into something a bit more python friendly.
    """
    def __init__(self, run, data_dir):
        self.run = run
        self.run_reader = FileReader(run, data_dir)
        self.extracted_values = dict()

    def __repr__(self):
        return ('<BeaconTau.RunAnalyzer for run ' + str(self.run) + '>')

    def split_attribute(self, attribute):
        attributes = attribute.split(':')
        return attributes

    def cached_extract(self, attribute):
        if attribute in self.extracted_values:
            return self.extracted_values[attribute]
        else:
            values = self.extract(attribute)
            if values is not None:
                self.extracted_values[attribute] = values
            return values

    def extract(self, attribute):

        values = None
        try:
            values = [s.__getattribute__(attribute) for s in self.run_reader.statuses]
            return values
        except:
            pass
        try:
            values = [h.__getattribute__(attribute) for h in self.run_reader.headers]
            return values
        except:
            pass
        try:
            values = [e.__getattribute__(attribute) for e in self.run_reader.events]
            return values
        except:
            pass

        raise AttributeError(attribute + ' is not something in BeaconTau.Status, BeaconTau.Header, or BeaconTau.Event!')
        return None

    def draw(self, attribute, show = False):
        plt.ion()
        plt.show()
        values = self.cached_extract(attribute)
        if values is not None:
            fig = plt.figure()
            mng = plt.get_current_fig_manager()
            mng.resize(*mng.window.maxsize())
            plt.xlabel('Entry')
            plt.ylabel(attribute)
            plt.title('Run ' + str(self.run))
            lines = plt.plot(values)
            labels = [attribute + '[' + str(c) + ']' for c in range(len(lines))]
            plt.legend(lines, labels)
            if show is True:
                plt.show()
            return fig

    def scan(self, attribute):
        values = self.cached_extract(attribute)
        if values is not None:
            entries = 0
            for entry, value in enumerate(values):
                to_print = str(entry) + '\t' + str(value)
                print(to_print)
                entries += 1
                if entry > 0 and (entry+1) % 25 == 0:
                    keys = input('Press q to quit: ')
                    if len(keys) > 0 and keys[0] == 'q':
                        break
            print('Finished scanning ' + str(entries) + ' entries')


    def get_entry(self, entry):
        return EventAnalyzer(self.run_reader.headers[entry],  self.run_reader.events[entry])

    def get_event(self, event_number):
        event_numbers = self.cached_extract('event_number')
        try:
            entry = event_numbers.index(event_number)
            return get_entry(entry)
        except:
            raise ValueError(str(event_number) + ' not found in run ' + str(self.run))
            return None



try:
    from setuptools import setup, Extension
    # Required for compatibility with pip (issue #177)
    from setuptools.command.install import install
except ImportError:
    from distutils.core import setup, Extension
    from distutils.command.install import install

from distutils.command.build import build
from distutils.command.clean import clean

import subprocess
import tempfile


subprocess.call(["pip", "install", "pybind11"])

import pybind11

pybind11_include_dir = pybind11.get_include()
libbeacon_clone_dir = tempfile.TemporaryDirectory()
libbeacon_version_tag = "0.1.4"


class BeaconTauClean(clean):
    def run(self):
        libbeacon_clone_dir.cleanup()
        clean.run(self)

class BeaconTauBuild(build):
    def run(self):
        # first we clone and build the required version of libbeacon.so
        clone_command = "git clone --branch v"+ libbeacon_version_tag + " https://github.com/beaconTau/libbeacon " + libbeacon_clone_dir.name
        clone_process = subprocess.Popen(clone_command, shell=True)
        clone_process.wait()
        # Then do the normal python building
        build.run(self)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="BeaconTau",
    version=libbeacon_version_tag, # try to keep these in sync
    author="Ben Strutt",
    author_email="strutt@physics.ucla.edu",
    description="A package for interacting with BEACON data in python",
    long_description=long_description,
    license='GPL3',
    url="https://github.com/beaconTau/BeaconTau",
    packages=['BeaconTau', 'BeaconTau/Flame'],
    ext_modules=[
        Extension('_BeaconTau', ['BeaconTau.cpp',  libbeacon_clone_dir.name + '/beacon.c'],
                  include_dirs = [libbeacon_clone_dir.name, pybind11_include_dir],
                  library_dirs = ['/usr/local/lib' ],
                  libraries=['z'],
                  extra_compile_args = ['-std=c++11']
                  )
    ],
    install_requires=[
        'pybind11', 'scipy', 'matplotlib', 'bokeh', 'numpy'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX",
    ],
    cmdclass={
        'build' : BeaconTauBuild,
        'clean' : BeaconTauClean,
    }
)
